# Copyrights --------------------------------------------------------------
# Author: Mirko Ledda
# Copyright 2020 Mirko Ledda under the BSD-2 License

# This script can be used to compute the pedigree-based additive relationship matrix (A),
# the genetic-based one (G) and to combine them into a single matrix (H).

rm(list=ls()); gc(verbose = T, reset = T, full = T)  # Cleanup and garbage collection

# Load required modules ---------------------------------------------------
library("data.table")
library("igraph")
library("synbreed")  # Package `regress` must be installed as a standalone
library("rrBLUP")
library("amap")
library("MASS")
library("Matrix") 

# IO ----------------------------------------------------------------------
pedigree_fp <- "pedigree.csv"
A_out <- "A.RData"  # A relationship matrix
G_out <- "G.RData"  # G relationship matrix
H_out <- "H.RData"  # H relationship matrix

# Functions ---------------------------------------------------------------
legarra_H <- function(A, G, scaleG=FALSE, isPD=FALSE) {
  # Combining matrices A and G using the method described in Legarra et al., 2009.
  # Full reference: 
  # Legarra, A., I. Aguilar, and I. Misztal. “A Relationship Matrix Including Full Pedigree and Genomic Information.” Journal of Dairy Science 92, no. 9 (September 2009): 4656–63. https://doi.org/10.3168/jds.2009-2061.
  # Method descibed under the "Modification for the Whole Pedigree" section
  
  # Inputs:
  #   A: Relationship matrix based on pedigree with named rows and columns
  #   G: Relationship matrix based on genomic data with named rows and columns
  #   scaled: Should G be scaled to match A prior to merging?
  #   isPD: Should the output H matrix be forced to be positive definite?
  
  require("MASS")
  require("Matrix") 
  
  # Match A and G
  A_ixs <- rownames(A)
  G_ixs <- rownames(G)
  
  rm_ixs <- G_ixs[!G_ixs %in% A_ixs]  # Remove ID that are not in A
  if (length(rm_ixs) > 0) {
    G_ixs <- G_ixs[!G_ixs %in% rm_ixs]
    G <- G[rownames(G) %in% G_ixs, colnames(G) %in% G_ixs]
  }
  
  A_flt <- !(A_ixs %in% G_ixs)
  G_flt <- A_ixs %in% G_ixs
  
  # Divide A
  A11 <- A[A_flt, A_flt]
  A12 <- A[A_flt, G_flt]
  A22 <- A[G_flt, G_flt]
  
  # Make sure A22 and G have the same index order
  reorder_ixs <- match(rownames(A22), rownames(G))
  G <- G[reorder_ixs, reorder_ixs]
  
  # Scale G to match the mean of diagonal and off-diagonal elements in A
  if (scaleG) {
    G <- scale_G(A22, G)
  }
  
  # Compute the needed submatrices to assemble H
  A22i <- ginv(as.matrix(A22))
  
  H11 <- A11 + A12 %*% A22i %*% (G-A22) %*% A22i %*% t(A12)
  H12 <- A12 %*% A22i %*% G
  H21 <- t(H12)
  H22 <- G
  
  # Assemble H
  H1 <- cbind(H11, H12)
  H2 <- cbind(H21, H22)
  H <- rbind(H1, H2)
  
  # This part here ensure positive definiteness of the H matrix and is not mentionned in the paper
  if (isPD) {
    if (!all(Re(eigen(H)$values) > 0)) {
      H <- nearPD(H)$mat
    }
  }
  
  return(H)
}

mean_offdiag <- function(X) {
  Xc <- X
  diag(Xc) <- NA
  return(mean(Xc, na.rm=T))
}

scale_G <- function(A22, G) {
  c <- mean(diag(G))
  a <- mean_offdiag(G)
  d <- mean(diag(A22))
  b <- mean_offdiag(A22)
  
  beta <- (d-b)/(c-a)
  alpha <- b-a*beta
  
  Gs <- G * beta + alpha
  
  return(Gs)
}

get_maf <- function(geno) {
  maf <- colMeans(geno, na.rm=T) / 2
  maf[maf>0.5] <- 1 - maf[maf>0.5]
  return(maf)
}

remove_maf <- function(geno, maf=0.05) {
  gmaf <- get_maf(geno)
  mask <- gmaf > maf
  geno <- geno[, mask]
  print(paste0("Removed ", sum(!mask), " SNPs based on MAF"))
  return(geno)
}

remove_missing <- function(geno, missingness=0.90) {
  mask <- colMeans(is.na(geno)) < missingness
  geno <- geno[, mask]
  print(paste0("Removed ", sum(!mask), " SNPs based on missingness"))
  
  mask <- rowMeans(is.na(geno)) < missingness
  geno <- geno[mask,]
  print(paste0("Removed ", sum(!mask), " individuals based on missingness"))
  
  return(geno)
}

LD_pruning <- function(geno, alpha=0.95, batch_size=1000, step_size=500, ncore=4) {
  # Rolling windows LD pruning based on hyerachical clustering to handle large files
  # This method is similar to what found in most common software packages such as PLINK
  
  batch_size <- min(ncol(geno), batch_size)
  rm_idx <- NULL
  batches = list()
  i <- 1
  j <- 1
  while (j < ncol(geno)) {
    k <- min(ncol(geno), j + step_size)
    batches[[i]] <- c(j, k)
    j <- j + step_size
    i <- i + 1
  }
  
  for (batch in batches) {
    g <- geno[, batch[1]:batch[2]]
    d <- Dist(t(g), method = "abspearson", nbproc=ncore)
    hc <- hclust(d, method = "complete")
    clusters <- cutree(hc, h=1-alpha)
    for (i in unique(clusters)) {
      rm_idx = c(rm_idx, names(clusters[clusters==i])[-1])
    }
  }
  
  mask = rep(T, ncol(geno))
  mask[colnames(geno) %in% rm_idx] <- F
  geno <- geno[, mask]
  
  print(paste0("Removed ", sum(!mask), " SNPs based on LD"))
  return(geno)
}

# Script starts here ------------------------------------------------------
data <- read.csv("pedigree_input.csv", header=T, fill=NA)

# Compute A ---------------------------------------------------------------
dped <- create.pedigree(
  as.character(data$Accession_ID), 
  as.character(data$Parent_1), 
  as.character(data$Parent_2),
  add.ancestors=T,
)
A <- kin(create.gpData(pedigree=dped), ret="add")
save(A, file=A_out)

# Compute G ---------------------------------------------------------------
load("istraw35.Rdata")
istraw35[istraw35 == -1] <- NA
geno <- istraw35

# Quality-based filterings
geno <- remove_missing(geno, missingness = 0.90)
geno <- remove_maf(geno, maf=0.05)
geno <- LD_pruning(geno, alpha=0.95, batch_size=1000, step_size=500, ncore=4)

# Compute G
M <- geno - 1  # [0, 1, 2] -> [-1, 0, 1] convention
G <- A.mat(M, min.MAF = 0.05, max.missing=0.90, impute.method="mean", n.core=4)
save(G, file=G_out)

# Compute H -----------------------------------------------------------------
H <- legarra_H(A, G, scaleG=TRUE, isPD=FALSE)
save(H, file=H_out)

# Unit test ---------------------------------------------------------------
# Testing against the "Example" section from Legarra et al., 2009
ped <- matrix(c(1, 0, 0, 
              2, 0, 0,
              3, 0, 0,
              4, 0, 0,
              5, 0, 0,
              6, 0, 0,
              7, 0, 0,
              8, 0, 0,
              9, 1, 2,
              10, 3, 4,
              11, 5, 6,
              12, 7, 8,
              13, 9, 10,
              14, 11, 12,
              15, 4, 11,
              16, 13, 15,
              17, 13, 14), 17, 3, byrow = T)
colnames(ped) <- c("ID", "P1", "P2")
ped <- as.data.frame(ped)

ped <- create.pedigree(
  as.character(ped$ID), 
  as.character(ped$P1), 
  as.character(ped$P2),
  add.ancestors=T,
)

A <- kin(ped, ret="add")

G <- matrix(0.7, 4, 4)
diag(G) <- 1
rownames(G) <- 9:12
colnames(G) <- 9:12

H <- legarra_H(A, G)
i <- match(1:17, colnames(H))
H <- H[i, i]

cat(" Published values:\n",
    "0.35 0.35 0.35 0.35 0.35 0.35 0.5 0.5 0.7 0.7 0.7 1 0.7 0.85 0.53 0.61 0.78\n",
    "Computed values:\n",
    round(H[12, ]+1E-6, 2)  # The +1E-6 is because 0.05 is rounded to 0.0 and not 0.1 by default
    )
